@@ -3,11 +3,13 @@ from __future__ import annotations
 from dataclasses import dataclass
 from typing import Any
 
+from geneticengine.core.problems import FitnessType
+
 
 @dataclass
 class Individual:
     genotype: Any
-    fitness: float | None = None
+    fitness: FitnessType | None = None
 
     def __str__(self) -> str:
         return str(self.genotype)
