@@ -274,7 +274,7 @@ def mk_save_init(starting_symbol: Any, receiver: Callable):
 def expand_node(
     r: Source,
     g: Grammar,
-    new_symbol,
+    new_symbol,  # Method to find new_symbol (?)
     filter_choices,
     receiver,
     depth,
@@ -345,16 +345,20 @@ def expand_node(
             )
 
         if starting_symbol in g.alternatives:  # Alternatives
+            extra_depth = 0
+            if is_abstract(starting_symbol) and g.expansion_depthing:
+                extra_depth = 1
+
             compatible_productions = g.alternatives[starting_symbol]
             valid_productions = filter_choices(
                 compatible_productions,
-                depth - 1,
+                depth - extra_depth,
             )
             if not valid_productions:
                 raise GeneticEngineError(
                     "No productions for non-terminal node with type: {} in depth {} (minimum required: {}).".format(
                         starting_symbol,
-                        depth,
+                        depth - extra_depth,
                         str(
                             [
                                 (vp, g.distanceToTerminal[vp])
@@ -372,7 +376,7 @@ def expand_node(
                 )
             else:
                 rule = r.choice(valid_productions, str(starting_symbol))
-            new_symbol(rule, receiver, depth - 1, id, ctx)
+            new_symbol(rule, receiver, depth - extra_depth, id, ctx)
         else:  # Normal production
             args = get_arguments(starting_symbol)
             ctx = ctx.copy()
@@ -417,11 +421,17 @@ def mutate_inner(
     i: TreeNode,
     max_depth: int,
     ty: type,
-    force_mutate: bool = False,
+    force_mutate: bool,
+    depth_aware_mut: bool,
 ) -> TreeNode:
-    if i.gengy_nodes > 0:
-        c = r.randint(0, i.gengy_nodes - 1)
-        if c == 0 or force_mutate:
+    counter = i.gengy_weighted_nodes if depth_aware_mut else i.gengy_nodes
+    if counter > 0:
+        c = r.randint(0, counter - 1)
+        if (
+            c == 0
+            or (c <= i.gengy_distance_to_term and depth_aware_mut)
+            or force_mutate
+        ):
             # If Metahandler mutation exists, the mutation process is different
             args_with_specific_mutation = [
                 has_annotated_mutation(arg[1]) for arg in get_arguments(i)
@@ -447,33 +457,50 @@ def mutate_inner(
                     method=Grow,
                     current_node=args[index],
                 )
-                return mk_save_init(type(i), lambda x: x)(*args)
+                mk = mk_save_init(type(i), lambda x: x)(*args)
+                return mk
 
-            replacement = random_node(r, g, max_depth, ty, method=Grow)
+            for _ in range(5):
+                replacement = random_node(r, g, max_depth, ty, method=Grow)
+                if replacement != i:
+                    break
             return replacement
         else:
-            if is_abstract(ty):
+            if is_abstract(ty) and g.expansion_depthing:
                 max_depth -= g.abstract_dist_to_t[ty][type(i)]
             max_depth -= 1
             args = list(i.gengy_init_values)
+            c -= i.gengy_distance_to_term if depth_aware_mut else 1
             for idx, (_, field_type) in enumerate(get_arguments(i)):
                 child = args[idx]
                 if hasattr(child, "gengy_nodes"):
-                    count = child.gengy_nodes
+                    count = (
+                        child.gengy_weighted_nodes
+                        if depth_aware_mut
+                        else child.gengy_nodes
+                    )
                     if c <= count:
-                        args[idx] = mutate_inner(
+                        mi = mutate_inner(
                             r,
                             g,
                             child,
                             max_depth,
                             field_type,
+                            force_mutate,
+                            depth_aware_mut,
                         )
+                        args[idx] = mi
                         break
                     else:
                         c -= count
-            return mk_save_init(i, lambda x: x)(*args)
+            mk = mk_save_init(i, lambda x: x)(*args)
+            return mk
     else:
-        return random_node(r, g, max_depth, ty, method=Grow)
+        for _ in range(5):
+            rn = random_node(r, g, max_depth, ty, method=Grow)
+            if rn != i:
+                break
+        return rn
 
 
 def mutate_specific_type_inner(
@@ -484,9 +511,18 @@ def mutate_specific_type_inner(
     ty: type,
     specific_type: type,
     n: int,
+    depth_aware_mut: bool,
 ) -> TreeNode:
     if n == 1 and type(i) == specific_type:
-        return mutate_inner(r, g, i, max_depth, ty, force_mutate=True)
+        return mutate_inner(
+            r,
+            g,
+            i,
+            max_depth,
+            ty,
+            force_mutate=True,
+            depth_aware_mut=depth_aware_mut,
+        )
     else:
         args = list(i.gengy_init_values)
         for idx, (_, field_type) in enumerate(get_arguments(i)):
@@ -503,6 +539,7 @@ def mutate_specific_type_inner(
                     ty,
                     specific_type,
                     n,
+                    depth_aware_mut,
                 )
             else:
                 n -= n_options
@@ -516,11 +553,20 @@ def mutate_specific_type(
     max_depth: int,
     target_type: type,
     specific_type: type,
+    depth_aware_mut: bool,
 ) -> TreeNode:
     ch = r.randint(0, 2)
     n_options = len(list(find_in_tree_exact(g, specific_type, i, max_depth)))
     if ch == 0 or n_options == 0:
-        new_tree = mutate_inner(r, g, i, max_depth, target_type)
+        new_tree = mutate_inner(
+            r,
+            g,
+            i,
+            max_depth,
+            target_type,
+            force_mutate=False,
+            depth_aware_mut=depth_aware_mut,
+        )
         relabeled_new_tree = relabel_nodes_of_trees(new_tree, g)
         return relabeled_new_tree
     else:
@@ -533,6 +579,7 @@ def mutate_specific_type(
             target_type,
             specific_type,
             n,
+            depth_aware_mut,
         )
         relabeled_new_tree = relabel_nodes_of_trees(new_tree, g)
         return relabeled_new_tree
@@ -545,10 +592,19 @@ def mutate(
     max_depth: int,
     target_type: type,
     specific_type: type | None = None,
+    depth_aware_mut: bool = False,
 ) -> Any:
     if specific_type:
-        return mutate_specific_type(r, g, i, max_depth, target_type, specific_type)
-    new_tree = mutate_inner(r, g, i, max_depth, target_type)
+        return mutate_specific_type(
+            r,
+            g,
+            i,
+            max_depth,
+            target_type,
+            specific_type,
+            depth_aware_mut,
+        )
+    new_tree = mutate_inner(r, g, i, max_depth, target_type, False, depth_aware_mut)
     relabeled_new_tree = relabel_nodes_of_trees(new_tree, g)
     return relabeled_new_tree
 
@@ -560,9 +616,9 @@ def find_in_tree(g: Grammar, ty: type, o: TreeNode, max_depth: int):
         for t in o.gengy_types_this_way:
 
             def is_valid(node):
-                _, depth, _ = relabel_nodes(node, g)
+                _, depth, _, _ = relabel_nodes(node, g)
 
-                if is_abs:
+                if is_abs and g.expansion_depthing:
                     depth += g.abstract_dist_to_t[ty][t]
 
                 return depth <= max_depth
@@ -580,7 +636,7 @@ def find_in_tree_exact(g: Grammar, ty: type, o: TreeNode, max_depth: int):
         if vals:
 
             def is_valid(node):
-                _, depth, _ = relabel_nodes(node, g)
+                _, depth, _, _ = relabel_nodes(node, g)
                 return depth <= max_depth
 
             yield from filter(is_valid, vals)
@@ -593,11 +649,17 @@ def crossover_inner(
     o: TreeNode,
     max_depth: int,
     ty: type,
-    force_crossover: bool = False,
+    force_crossover: bool,
+    depth_aware_co: bool,
 ) -> Any:
-    if i.gengy_nodes > 0:
-        c = r.randint(0, i.gengy_nodes - 1)
-        if c == 0 or force_crossover:
+    counter = i.gengy_weighted_nodes if depth_aware_co else i.gengy_nodes
+    if counter > 0:
+        c = r.randint(0, counter - 1)
+        if (
+            c == 0
+            or (c <= i.gengy_distance_to_term and depth_aware_co)
+            or force_crossover
+        ):
             replacement = None
             args_with_specific_crossover = [
                 has_annotated_crossover(arg[1]) for arg in get_arguments(i)
@@ -636,18 +698,26 @@ def crossover_inner(
             if options:
                 replacement = r.choice(options)
             if replacement is None:
-                replacement = random_node(r, g, max_depth, ty, method=Grow)
+                for _ in range(5):
+                    replacement = random_node(r, g, max_depth, ty, method=Grow)
+                    if replacement != i:
+                        break
 
             return replacement
         else:
-            if is_abstract(ty):
+            if is_abstract(ty) and g.expansion_depthing:
                 max_depth -= g.abstract_dist_to_t[ty][type(i)]
             max_depth -= 1
             args = list(i.gengy_init_values)
+            c -= i.gengy_distance_to_term if depth_aware_co else 1
             for idx, (field, field_type) in enumerate(get_arguments(i)):
                 child = args[idx]
                 if hasattr(child, "gengy_nodes"):
-                    count = child.gengy_nodes
+                    count = (
+                        child.gengy_weighted_nodes
+                        if depth_aware_co
+                        else child.gengy_nodes
+                    )
                     if c <= count:
                         args[idx] = crossover_inner(
                             r,
@@ -656,6 +726,8 @@ def crossover_inner(
                             o,
                             max_depth,
                             field_type,
+                            force_crossover=False,
+                            depth_aware_co=depth_aware_co,
                         )
                         break
                     else:
@@ -674,9 +746,19 @@ def crossover_specific_type_inner(
     ty: type,
     specific_type: type,
     n: int,
+    depth_aware_co: bool,
 ) -> TreeNode:
     if n == 1 and type(i) == specific_type:
-        return crossover_inner(r, g, i, o, max_depth, ty, force_crossover=True)
+        return crossover_inner(
+            r,
+            g,
+            i,
+            o,
+            max_depth,
+            ty,
+            force_crossover=True,
+            depth_aware_co=depth_aware_co,
+        )
     else:
         args = list(i.gengy_init_values)
         for idx, (_, field_type) in enumerate(get_arguments(i)):
@@ -694,6 +776,7 @@ def crossover_specific_type_inner(
                     ty,
                     specific_type,
                     n,
+                    depth_aware_co=depth_aware_co,
                 )
             else:
                 n -= n_options
@@ -708,12 +791,22 @@ def crossover_specific_type(
     max_depth: int,
     target_type: type,
     specific_type: type,
+    depth_aware_co: bool,
 ) -> TreeNode:
     ch = r.randint(0, 1)
     n_options_i = len(list(find_in_tree_exact(g, specific_type, i, max_depth)))
     n_options_o = len(list(find_in_tree_exact(g, specific_type, o, max_depth)))
     if ch == 0 or n_options_i == 0 or n_options_o == 0:
-        new_tree = crossover_inner(r, g, i, o, max_depth, target_type)
+        new_tree = crossover_inner(
+            r,
+            g,
+            i,
+            o,
+            max_depth,
+            target_type,
+            force_crossover=False,
+            depth_aware_co=depth_aware_co,
+        )
         relabeled_new_tree = relabel_nodes_of_trees(new_tree, g)
         return relabeled_new_tree
     else:
@@ -727,6 +820,7 @@ def crossover_specific_type(
             target_type,
             specific_type,
             n,
+            depth_aware_co=depth_aware_co,
         )
         relabeled_new_tree = relabel_nodes_of_trees(new_tree, g)
         return relabeled_new_tree
@@ -739,6 +833,7 @@ def crossover(
     p2: TreeNode,
     max_depth: int,
     specific_type: type | None = None,
+    depth_aware_co: bool = False,
 ) -> tuple[TreeNode, TreeNode]:
     """
     Given the two input trees [p1] and [p2], the grammar and the random source, this function returns two trees that are created by crossing over [p1] and [p2]. The first tree returned has [p1] as the base, and the second tree has [p2] as a base.
@@ -752,9 +847,19 @@ def crossover(
             max_depth,
             g.starting_symbol,
             specific_type,
+            depth_aware_co=depth_aware_co,
         )
     else:
-        new_tree1 = crossover_inner(r, g, p1, p2, max_depth, g.starting_symbol)
+        new_tree1 = crossover_inner(
+            r,
+            g,
+            p1,
+            p2,
+            max_depth,
+            g.starting_symbol,
+            force_crossover=False,
+            depth_aware_co=depth_aware_co,
+        )
     relabeled_new_tree1 = relabel_nodes_of_trees(new_tree1, g)
 
     if specific_type:
@@ -766,9 +871,19 @@ def crossover(
             max_depth,
             g.starting_symbol,
             specific_type,
+            depth_aware_co=depth_aware_co,
         )
     else:
-        new_tree2 = crossover_inner(r, g, p2, p1, max_depth, g.starting_symbol)
+        new_tree2 = crossover_inner(
+            r,
+            g,
+            p2,
+            p1,
+            max_depth,
+            g.starting_symbol,
+            force_crossover=False,
+            depth_aware_co=depth_aware_co,
+        )
     relabeled_new_tree2 = relabel_nodes_of_trees(new_tree2, g)
     return relabeled_new_tree1, relabeled_new_tree2
 
@@ -785,8 +900,10 @@ class TreeBasedRepresentation(Representation[TreeNode]):
         depth: int,
         ty: type,
         specific_type: type = None,
+        depth_aware_mut: bool = False,
     ) -> TreeNode:
-        return mutate(r, g, ind, depth, ty, specific_type)
+        new_ind = mutate(r, g, ind, depth, ty, specific_type, depth_aware_mut)
+        return new_ind
 
     def crossover_individuals(
         self,
@@ -796,8 +913,9 @@ class TreeBasedRepresentation(Representation[TreeNode]):
         i2: TreeNode,
         max_depth: int,
         specific_type: type = None,
+        depth_aware_co: bool = False,
     ) -> tuple[TreeNode, TreeNode]:
-        return crossover(r, g, i1, i2, max_depth, specific_type)
+        return crossover(r, g, i1, i2, max_depth, specific_type, depth_aware_co)
 
     def genotype_to_phenotype(self, g: Grammar, genotype: TreeNode) -> TreeNode:
         return genotype
